import logging
import sys
import socket

from logging import CRITICAL, DEBUG, ERROR, INFO, WARNING
from logging.handlers import SysLogHandler
from collections import namedtuple
from typing import Union, Optional

from format_cef import format_cef
from format_cef.cef import CefValueError

from typeguard import typechecked

logger = logging.getLogger(__name__)

_levels = namedtuple("logging_levels", ["critical", "debug", "error", "info", "warning"])
logging_levels = _levels(CRITICAL, DEBUG, ERROR, INFO, WARNING)

SYSLOG_FACILITY = SysLogHandler.LOG_SYSLOG
VENDOR = "Osirium"
PRODUCT = "PPA"


class InvalidCEFName(Exception):
    pass


class InvalidCEFValue(Exception):
    pass


@typechecked
def stdout_logger(level: int = logging_levels.info):
    """
    Returns a logger that can send log messages to the PPA task logs window.

    :param level: Optional. Sets the threshold for the returned logger to log. Any logging messages which are less
        severe than this level will be ignored. The available log levels that can be used can be found here:
        :ref:`loggers:Logging Levels`

    :returns: A standard output logger

    Example 1 - Logging messages at info level to the PPA logs window

    .. code-block:: python

        from loggers import logging_levels, stdout_logger

        # Create basic stdout logger with default arguments
        logger = stdout_logger()
        logger.info("This is an info log message")

    Example 2 - Logging messages at debug level to the PPA logs window

        # Create a stdout logger at debug level
        logger = stdout_logger(level=logging_levels.debug)

        # This will not be logged
        logger.info("This is an info log message")

        # This will be logged
        logger.debug("This is a debug log message")

    """
    logging.basicConfig(
        stream=sys.stdout,
        level=level,
        format="%(asctime)s %(module)s.%(funcName)s:%(lineno)d %(levelname)s: %(message)s",
    )
    return logging.getLogger(__name__)


@typechecked
def syslog_logger(
    syslog_address: str, syslog_port: Union[str, int] = 514, level: int = logging_levels.info, use_tcp: bool = False
):
    """
     Returns a logger that can send syslog events to the supplied server details.

    :param syslog_address: The IP address or hostname of the target syslog server.
    :param syslog_port: Optional. Defaults to 514. The port the syslog server is listening on.
    :param level: Optional. Sets the threshold for the returned logger to log. Any logging messages which are less
        severe than this level will be ignored. The available log levels that can be used can be found here:
        :ref:`loggers:Logging Levels`
    :param use_tcp: Optional. Defaults to `False`. When set to `True` syslog messages will be sent over TCP. When set
        to `False` they will be sent over UDP.

    :returns: A syslog logger

    Example 1 - Sending plain text syslog messages:

    .. code-block:: python

        from loggers import syslog_logger, logging_levels

        # Create syslog logger
        logger = syslog_logger(address="10.10.1.12", port=5000)
        logger.info("This is an info-level syslog message")

    Example 2 - Sending CEF formatted syslog messages over TCP:

    .. code-block:: python

        from loggers import syslog_logger, format_cef_syslog_message

        # Create syslog logger
        logger = syslog_logger(address="10.10.1.12", port=5000, use_tcp=True)

        cef_message = format_cef_syslog_message(
            event_id=12,
            event_name="User Creation",
            severity=5,
            event_metadata={
                "agentAddress":"10.10.1.2"
            }
        )

        logger.info(cef_message)
    """
    sys_logger = logging.getLogger(__name__)
    sys_logger.addHandler(
        SysLogHandler(
            address=(syslog_address, int(syslog_port)),
            facility=SYSLOG_FACILITY,
            socktype=socket.SOCK_STREAM if use_tcp else socket.SOCK_DGRAM,
        )
    )
    sys_logger.setLevel(level)

    return sys_logger


@typechecked
def format_cef_syslog_message(
    *,
    event_id: Union[int, str],
    event_name: str,
    severity: int,
    product_version: Union[int, str, float] = "2.3",
    event_metadata: Optional[dict] = None,
):
    """
    Produces a CEF compliant syslog message using the supplied arguments and event metadata.

    :param event_id: A unique identifier for the type of event being reported.
    :param event_name: A description of the event being reported.
    :param severity: A severity of the event being report. **Must** be between 0 and 10.
    :param product_version: Optional. Defaults to 2.3. The product version to be used in the CEF log message.
    :param event_metadata: Optional. A dictionary of additional event metadata, the keys in the dictionary
            **must** be one of the Event Metadata Key in the table shown here: :ref:`cef_extensions:CEF Extensions`

    :raises InvalidCEFValue: if the value for a key in `event_metadata` is not the expected value.
    :raises InvalidCEFName: if the `event_metadata` key is not a valid CEF extension.

    :returns: A CEF compliant syslog message


    Example:

     .. code-block:: python

        cef_message = loggers.format_cef_syslog_message(
            event_id=12,
            event_name="User Creation",
            severity=5,
            event_metadata={
                "agentAddress":"10.10.1.2"
            }
        )

        print(cef_message)

        CEF:0|Osirium|PPA|2.3|12|User Creation|5|agt=10.10.1.2

    """
    if severity < 0 or severity > 10:
        raise InvalidCEFValue("Message severity must be greater than or equal to 0 and no greater than 10.")
    try:
        return format_cef(
            VENDOR,
            PRODUCT,
            str(product_version),
            str(event_id),
            event_name,
            severity,
            event_metadata if event_metadata else {},
        ).decode("utf-8")
    except KeyError as e:
        raise InvalidCEFName(
            f"{e} must be a valid CEF extension. Please view the documentation for valid CEF extensions."
        )
    except CefValueError as e:
        logger.error("Failed to format syslog message, please view the logs below:")
        logger.exception(e)
        raise InvalidCEFValue(
            "An invalid value was provided in the event metadata. Please view the task log for more information."
        )
