from typing import Dict, Union

from validators import mac_address as mac_validator

from ppa import ui


class InvalidDelimiter(Exception):
    pass


class InvalidMacAddress(Exception):
    pass


class InvalidFormat(Exception):
    pass


MAC_ADDRESS_DELIMITERS: Dict[str, Dict] = {
    ":": {"replace": lambda mac, new_delimiter: mac.replace(":", new_delimiter), "format": "xx:xx:xx:xx:xx:xx"},
    "-": {"replace": lambda mac, new_delimiter: mac.replace("-", new_delimiter), "format": "xx-xx-xx-xx-xx-xx"},
    "": {
        "replace": lambda mac, new_delimiter: "".join(
            f"{char}{new_delimiter}" if i < len(mac) and i % 2 == 0 else char for i, char in enumerate(mac, 1)
        )
        if len(mac) == 12
        else mac,
        "format": "xxxxxxxxxxxx",
    },
}


def format_mac_address(mac: str, new_delimiter: str = ":") -> str:
    """ Format a MAC address with a new delimiter.

    :param mac: The MAC address to format.
    :param new_delimiter: The delimiter to use. Supported delimiters are `:`, `-`, and
        no delimiter.

    :raises InvalidDelimiter: if the supplied delimiter is not supported.
    :raises InvalidFormat: if the supplied MAC address format is not supported.
    :returns: A formatted MAC address.
    """
    if not isinstance(mac, str):
        raise InvalidFormat(
            f"Value '{mac}' uses unsupported MAC address format. Supported MAC formats: "
            f"{[MAC_ADDRESS_DELIMITERS[key]['format'] for key in MAC_ADDRESS_DELIMITERS.keys()]}"
        )

    try:
        MAC_ADDRESS_DELIMITERS[new_delimiter]
    except KeyError:
        raise InvalidDelimiter(
            f"Supplied delimiter must be one of the following: {list(MAC_ADDRESS_DELIMITERS.keys())}"
        )

    for delimiter in MAC_ADDRESS_DELIMITERS.keys():
        mac = MAC_ADDRESS_DELIMITERS[delimiter]["replace"](mac, new_delimiter)

    if len(mac) in {12, 17}:
        return mac

    raise InvalidFormat(
        f"Value '{mac}' uses unsupported MAC address format. Supported MAC formats: "
        f"{[MAC_ADDRESS_DELIMITERS[key]['format'] for key in MAC_ADDRESS_DELIMITERS.keys()]}"
    )


def validate_mac_address(mac: str) -> bool:
    """ Validate a MAC address. The MAC address can be include delimiters `:`, `-`, or no delimiter.

    :param mac: The MAC address to validate.

    :raises InvalidFormat: if the supplied MAC address format is not supported.

    :returns: True if the MAC address is valid. False if the MAC address is not valid.
    """
    mac: str = format_mac_address(mac, new_delimiter=":")  # type: ignore
    try:
        assert mac_validator(mac)
        return True
    except AssertionError:
        return False


def mac_address(text: str, delimiter: str = ":", output_id: Union[str, None] = None) -> str:
    """ Prompt the user to input a MAC address. Validate the MAC address, format
    it using a new delimiter if supplied, and return it to the PPA task.

    :param text: The text to display with the input field.
    :param delimiter: The delimiter to use in the returned MAC address. Defaults to `:`.
    :param output_id: The element ID to use for the input field in the PPA interface.

    :returns: A validated and formatted MAC address.
    """
    while True:
        try:
            mac = ui.input_text(text, required=True, element_id=output_id)

            if validate_mac_address(mac):
                return format_mac_address(mac, delimiter)
            else:
                raise InvalidMacAddress(f"Provided Mac Address is invalid")
        except (InvalidFormat, InvalidMacAddress) as e:
            ui.output_error(str(e))
