"""
Excel

This module contains actions used to create Microsoft Excel workbooks.
"""
import logging
import pathlib

from typing import List, Dict, Any, Union

from openpyxl import Workbook

from ppa import ui


logger = logging.getLogger(__name__)

Sheet = List[Dict[str, Any]]
MultipleSheets = Dict[str, Sheet]
SHARED_WORKSPACE = pathlib.Path("/workspace/share")


def create_from_dictionaries(header: List[str], keys: List[str], data: Union[Sheet, MultipleSheets], file_name: str) -> None:
    """Generate an Excel workbook & save it as a file in the task workspace.

Supports creating workbooks with a single sheet _or_ multiple sheets.

The file will be saved using the **xlsx** extension.

!!! question "Minimum Plugin Version: 5.0.0"

??? info "Input (Single Sheet)"

    - __header:__ a list of column names

    - __keys:__ a list of dictionary keys used to retrieve the value for each column

    - __data:__ a list of dictionaries used to populate each row in the spreadsheet

    - __file_name:__ the name of the file to create (an **xlsx** extension will be added if not provided)

??? info "Input (Multiple Sheets)"

    - __header:__ a list of column names

    - __keys:__ a list of dictionary keys used to retrieve the value for each column

    - __data:__ a dictionary containing sheet names & dictionaries used to populate each sheet

    - __file_name:__ the name of the file to create (an **xlsx** extension will be added if not provided)

??? success "Output"

    Nothing is outputted by this action.

??? example "Example 1 - Single Sheet"

    - Getting Remote Desktop Users from Active Directory & saving them as `rdp_users`

    ```YAML linenums="1" hl_lines="5"
    - active_directory.groups.get_users:
        distinguishedName: CN=Remote Desktop Users,CN=Builtin,DC=Example,DC=Domain
      load:
        domain_controller: domain_controller
      save: users
    ```

    - Creating an Excel spreadsheet from `users` with 4 attributes from each user

    ```YAML linenums="1" hl_lines="12 14"
    - ppa_tools.excel.create_from_dictionaries:
        header:
          - SAM Account Name
          - Email Address
          - Enabled
          - Locked
        keys:
          - sAMAccountName
          - mail
          - is_enabled
          - is_locked
        file_name: rdp_users.xlsx
      load:
        data: users
    ```

    - Sending the created file as an email attachment

    {% raw %}
    ```YAML linenums="1" hl_lines="9"
    - ppa.events.send_email:
        subject: Test Excel Spreadsheet
        html: >
          <html>
            <head>PPA Email Header</head>
            <body>PPA Email Body</body>
          </html>
        attachments:
          - rdp_users.xlsx
    ```
    {% endraw %}

??? example "Example 2 - Multiple Sheets"

    - Getting users from 3 groups in Active Directory

    - Creating an Excel spreadsheet with a sheet per group & 4 attributes per user

    - Sending the spreadsheet as an email attachment

    {% raw %}
    ```YAML linenums="1"
    set:
      group_names:
        - Remote Desktop Users
        - Domain Admins
        - Account Operators
      group_members: {}

    steps:

      - name: Get Group Members
        sequence: group_names  # Loop over each group name defined above
        actions:

          # Look up the group using its sAMAccountName
          - active_directory.groups.by_samaccountname:
            load:
              sAMAccountName: loop.step.item.value
              domain_controller: domain_controller
            save: group

          # Get the users in the group
          - active_directory.groups.get_users:
            load:
              distinguishedName: group.distinguishedName
              domain_controller: domain_controller
            save: users

          # Update the group_members dictionary with the current group & its users
          - ppa_tools.dictionaries.insert:
            load:
              name: loop.step.item.value
              value: users
              dictionary: group_members
            save: group_members

      - name: Create & Send Spreadsheet
        actions:

        - ppa_tools.excel.create_from_dictionaries:
            header:
              - SAM Account Name
              - Email Address
              - Enabled
              - Locked
            keys:
              - sAMAccountName
              - mail
              - is_enabled
              - is_locked
            file_name: all_users
          load:
            data: group_members
          save: file

      - ppa.events.send_email:
          subject: Test Excel Spreadsheet
          html: >
            <html>
              <head>PPA Email Header</head>
              <body>PPA Email Body</body>
            </html>
          attachments:
            - "{{ spreadsheet_file_name }}"
    ```
    {% endraw %}

!!! tip "Header & Keys Length"

    The length of the `header` & `keys` inputs must be the same.
    """
    file_name = f"{file_name}.xlsx" if not file_name.endswith(".xlsx") else file_name
    destination_path = SHARED_WORKSPACE / file_name

    if len(header) != len(keys):
        message = (
            f"The length of the 'header' & 'keys' inputs must be the same. " f"Received {len(header)} headers & {len(keys)} keys."
        )
        ui.output_error(message)
        raise ValueError(message)

    if not any([isinstance(data, list), isinstance(data, dict)]):
        message = f"The 'data' input must be a list or a dictionary. Received {type(data)}."
        ui.output_error(message)
        raise ValueError(message)

    def _get_cell_value(record, key):
        return record.get(key) if record.get(key) is not None else ""

    def _process_rows(record):
        return [_get_cell_value(record, key) for key in keys]

    def _write_rows(record_list, sheet):
        list(map(lambda record: sheet.append(_process_rows(record)), record_list))

    workbook = Workbook()

    if isinstance(data, list):
        sheet = workbook.active
        sheet.append(header)
        _write_rows(data, sheet)
    else:
        for i, (sheet_name, records) in enumerate(data.items()):
            if i == 0:  # Use the default active sheet the first time round.
                sheet = workbook.active
                sheet.title = sheet_name
            else:  # All subsequent sheets are created explicitly.
                sheet = workbook.create_sheet(title=sheet_name)
            sheet.append(header)
            _write_rows(records, sheet)

    logger.info(f"writing excel spreadsheet to {destination_path}")
    workbook.save(filename=destination_path)
