"""
Password Generators

This module contains actions used to generate different types of password.
"""

import uuid
import logging
import random
import json
import re
import pkg_resources

from typing import List, Optional, TypedDict, Generator, Dict

import er

from ppa import ui


logger = logging.getLogger(__name__)


class InvalidRegex(Exception):
    pass


class DictionaryPassword(TypedDict):
    """
- Contains a password generated using a combination of dictionary words

- Several formats of the password are available to use


!!! info "Password Formats"

    The __words__ key contains a list of each word in the password in _lower_ case.

    The following keys contain the password with no spaces:

    - __lower__

    - __upper__

    - __mixed__

    - __title__

    The following keys contain the password with a space between each word:

    - __lower_spaced__

    - __upper_spaced__

    - __mixed_spaced__

    - __title_spaced__

    See below for more information on how spaces affect password length.


!!! warning "Spaces & Password Length"

    Formats that include spaces will have the spaces added _after_ the password is generated.

    This makes these formats longer than the supplied password length.

    If your requirement is to use passwords with spaces, adjust the __word_count__ & __length__ inputs to suit.

    The snippet below will generate a password whose _spaced_ formats are 32 characters in length:

    ```YAML linenums="1"
      - ppa_tools.password_generators.dictionary_words:
          length: 27
          word_count: 6
        save: password
    ```
    """

    lower: str
    upper: str
    mixed: str
    title: str
    lower_spaced: str
    upper_spaced: str
    mixed_spaced: str
    title_spaced: str
    words: List[str]


class PolicyCharacterSets(TypedDict):
    lower: str
    upper: str
    digit: str
    symbol: str


class PasswordPolicy(TypedDict):
    """
- Contains a character policy used to generate passwords

- At least one of the keys below must be supplied

!!! info "PasswordPolicy Keys"

    - __lower__: number of lower case letters

    - __upper__: number of upper case letters

    - __digit__: number of digits (0-9)

    - __symbol__: number of special characters (taken from `! # $ % & ^ * : ; ?`)
    """

    lower: Optional[int] = None
    upper: Optional[int] = None
    digit: Optional[int] = None
    symbol: Optional[int] = None


def _validate_policy(policy: PasswordPolicy):
    try:
        policy = {k: int(v) for k, v in policy.items() if v is not None}
    except (TypeError, ValueError):
        ui.output_error("The supplied policy contains a value that is not a number.")
        raise TypeError(f"Invalid password policy: {policy}")

    if policy == {}:
        message = "Cannot generate a password as no values were provided in the policy."
        ui.output_error(message)
        raise ValueError(message)

    if sum(policy.values()) == 0:
        message = "The supplied policy describes an empty password."
        ui.output_error(message)
        raise ValueError(message)
    return policy


def _get_word_length_combinations(
    *, valid_word_lengths: List[int], character_target: int, word_count: int = 3, current_numbers=[]
) -> Generator[list, None, None]:
    current_sum = sum(current_numbers)
    if len(current_numbers) > word_count:
        return
    if current_sum == character_target and len(current_numbers) == word_count:
        yield current_numbers
    if current_sum >= character_target:
        return
    for number in valid_word_lengths:
        yield from _get_word_length_combinations(
            valid_word_lengths=valid_word_lengths,
            character_target=character_target,
            current_numbers=current_numbers + [number],
            word_count=word_count,
        )


def _password_from_words(*, length: int = 15, word_count: int = 3, valid_word_lengths: List[int]) -> Dict[str, str]:
    logger.debug(
        f"creating english word password with length {length}, "
        f"word_count {word_count}, and valid_word_lengths {sorted(valid_word_lengths)}."
    )

    if length not in list(range(1, 29)):
        raise ValueError("Password length must be 1 - 28.")
    if word_count not in list(range(1, 11)):
        raise ValueError("Word count must be 1 - 10.")

    with open(pkg_resources.resource_filename("ppa_tools", "config/words.json"), "r") as open_file:
        words = json.loads("\n".join(open_file.readlines()))

    if all([length > 20, word_count <= 2]):
        # The longest word is 10 characters to make them easier to remember.
        raise ValueError("Cannot create a password longer than 20 characters using no more than 2 words.")

    # Randomise the list of ints before feeding them to the algorithm, making the
    # generator return a different order of results every time. This means we can
    # take the first value from the generator and still have it be different every time.
    random.shuffle(valid_word_lengths)

    try:
        word_lengths = next(
            _get_word_length_combinations(valid_word_lengths=valid_word_lengths, character_target=length, word_count=word_count)
        )
        words = [random.choice(words[str(word_length)]) for word_length in word_lengths]
        random.shuffle(words)

        words_lower = [w.lower() for w in words]
        words_upper = [w.upper() for w in words]
        words_title = [w.title() for w in words]

        # Decide which characters of each word should be converted to upper case when generating the mixed
        # case passwords. This is locked in early so both the regular and spaced formats of mixed case
        # passwords have the same casing pattern. Only ~half of each word is converted to upper case.
        upper_case_indices = [list(map(lambda _: random.randint(0, len(word)), range(0, int(len(word) / 2)))) for word in words]

        # Now generate a new list of words where each word has had the random upper casing applied.
        words_mixed = [
            "".join([char.upper() if i in upper_case_indices[x] else char for i, char in enumerate(word, 0)])
            for x, word in enumerate(words, 0)
        ]

        delimiters = ["", " "]
        lower, lower_spaced = [delim.join(words_lower) for delim in delimiters]
        upper, upper_spaced = [delim.join(words_upper) for delim in delimiters]
        title, title_spaced = [delim.join(words_title) for delim in delimiters]
        mixed, mixed_spaced = [delim.join(words_mixed) for delim in delimiters]

        return {
            "words": words,
            "lower": lower,
            "upper": upper,
            "title": title,
            "mixed": mixed,
            "lower_spaced": lower_spaced,
            "upper_spaced": upper_spaced,
            "title_spaced": title_spaced,
            "mixed_spaced": mixed_spaced,
        }
    except (StopIteration, IndexError):
        raise ValueError(
            "The supplied password configuration is either not compatible "
            "with the word list, or describes an impossible password."
        )


def regex(regex: str = r"[A-Za-z0-9\-£\:\$\#\%\&\^]", length: int = 24) -> str:
    r"""Generate a password based on a regular expression.

    !!! question "Minimum Plugin Version: 6.0.0"

    ??? info "Input"

        - __regex:__ a regular expression containing a set of characters (defaults to _[A-Za-z0-9\-£\:\$\#\%\&\^]_)

        - __length:__ the length of the generated password (defaults to _15_)

    ??? success "Output"
        The generated password as a string.

    ??? example

        Generating a 20 character password using the default character set.

        ```YAML linenums="1"
        - ppa_tools.generate.regex_password:
            length: 20
          save: password
        ```
    """
    try:
        return er.generate(f"{regex}{{{length}}}", shuffle=True)
    except re.error as e:
        raise InvalidRegex(f"Failed to compile supplied regex {regex}: {str(e)}")


def from_policy(policy: PasswordPolicy) -> str:
    """Generate a password that meets the supplied [PasswordPolicy](../inputs_outputs/#passwordpolicy).

    !!! question "Minimum Plugin Version: 6.2.0"

    ??? info "Input"

        - __policy:__ a [PasswordPolicy](../inputs_outputs/#passwordpolicy) used to generate the password

    ??? success "Output"

        The generated password as a string.

    ??? example

        ```YAML linenums="1"
        - ppa_tools.generate.from_policy:
            policy:
              lower: 5
              upper: 5
              digit: 5
              symbol: 5
          save: password
        ```
    """
    policy = _validate_policy(policy)
    PolicyCharacters = PolicyCharacterSets(lower="[a-z]", upper="[A-Z]", digit="[0-9]", symbol="[!#$%&^*:;?]")
    password_chunks = [er.generate(f"{PolicyCharacters[character_set]}{{{count}}}") for character_set, count in policy.items()]
    password_string = "".join(password_chunks)
    password_list = list(password_string)
    random.shuffle(password_list)
    return "".join(password_list)


def dictionary_words(
    *, length: int = 15, word_count: int = 3, valid_word_lengths: Optional[List[int]] = None
) -> DictionaryPassword:
    """Generate a password containing 1 or more dictionary words.

    Multiple formats of the password will be generated.

    See [here](../inputs_outputs/#dictionarypassword) for the list of formats, & important information on formats that include spaces.

    !!! question "Minimum Plugin Version: 6.0.0"

    ??? info "Input"

        - __length:__ the length of the generated password (defaults to _15_)

        - __word_count:__ the number of words the password should consist of (defaults to _3_)

        - __valid_word_lengths:__ a list of valid word lengths (defaults to _3_-_8_)

    ??? success "Output"

        A [DictionaryPassword](../inputs_outputs/#dictionarypassword).

    ??? example

        Generating a 20 character password consisting of 4 words.

        ```YAML linenums="1"
        - ppa_tools.password_generators.dictionary_words:
            length: 20
            word_count: 4
          save: password
        ```
    """
    return _password_from_words(length=length, word_count=word_count, valid_word_lengths=valid_word_lengths or list(range(3, 9)))


def dictionary_words_interactive(
    *,
    length: int = 15,
    word_count: int = 3,
    valid_word_lengths: Optional[List[int]] = None,
    format: Optional[str] = None,
    title: Optional[str] = None,
) -> str:
    """Interactively generate a password containing 1 or more dictionary words.

    Each time a password is generated, the Task Operator will be given the choice to:

    - _Use Password_

    - _Generate Another_

    The process can be repeated as many times as needed.

    This action will output the password in one of the formats listed [here](../inputs_outputs/#dictionarypassword).

    !!! question "Minimum Plugin Version: 6.0.0"

    !!! danger "Remember"

        Each generated password will be displayed to the Task Operator.

    ??? info "Input"

        - __length:__ the length of the generated password (defaults to `15`)

        - __word_count:__ the number of words the password should consist of (defaults to `3`)

        - __valid_word_lengths:__ a list of valid word lengths (defaults to `3`-`8`)

        - __title:__ a title displayed to the Task Operator (defaults to `Generated Password`)

        - __format:__ a [DictionaryPassword](../inputs_outputs/#dictionarypassword) format to output (defaults to __lower__)

    ??? success "Output"

        The password in the format specified by the __format__ input.

    ??? example "Upper Case"

        Generating an upper case password using the default __word_count__ & __length__:

        ```YAML linenums="1"
        - ppa_tools.password_generators.dictionary_words_interactive:
            format: upper
          save: password
        ```

    ??? example "Lower Case With Spaces"

        Generating a 32 character lower case password that includes spaces.

        Because spaces are added _after_ generation, the password length is 27 with a word count of 6.

        This is one example combination that yields a 32 character password:

        ```YAML linenums="1"
        - ppa_tools.password_generators.dictionary_words_interactive:
            length: 27
            word_count: 6
            format: lower_spaced
          save: password
        ```
    """
    format = (format or "lower").lower()
    valid_formats = list(DictionaryPassword.__annotations__.keys())

    if format not in valid_formats:
        ui.output_error(f"Supplied password format '{format}' is not supported.")
        format = (
            ui.input_choice("Please select a password format", options=[f"{f.replace('_', ' ').title()}" for f in valid_formats])
            .replace(" ", "_")
            .lower()
        )

    markdown_id = str(uuid.uuid4())
    confirm_id = str(uuid.uuid4())
    while True:
        generated_password_dict = _password_from_words(
            length=length, word_count=word_count, valid_word_lengths=valid_word_lengths or list(range(3, 9))
        )
        password = generated_password_dict[format]
        html_lines = [
            "<html>",
            "<body>",
            "<div style='font-weight: bold; float: left;'>Password: </div>",
            f"<div style='padding-left: 5px; font-style: bold; float: left;'>{password}</div>",
        ]
        # Show the password in a more readable format if the selected format isn't very readable.
        if format not in ["title_spaced", "upper_spaced", "lower_spaced"]:
            html_lines.extend(
                [
                    "<div style='float: left; padding-left: 10px;'>(Readable Version:</div>",
                    f"<div style='color: orange; font-style: italic; float: left;'>&nbsp;{generated_password_dict['title_spaced']}</div>",
                    "<div style='float: left;'>)</div>",
                ]
            )
        html_lines.extend(["</body>", "</html>"])

        markdown = """
### {title}

{html}
        """.format(
            title=title or "Generated Password", html="\n".join(html_lines),
        )

        ui.output_markdown(markdown, element_id=markdown_id)
        if ui.input_confirm("Please choose an option", confirm="Use Password", cancel="Generate Another", element_id=confirm_id):
            return password
