"""
Text

This module contains actions used to process & search text.
"""

import re
import logging


from typing import List, Dict, TypedDict, Union, Optional


from ppa import ui


logger = logging.getLogger(__name__)


class InvalidExpression(ValueError):
    pass


class AmbiguousExpression(ValueError):
    pass


def _to_int(value):
    try:
        return int(value)
    except (ValueError, TypeError):
        message = f"Supplied line number {value} is not a number."
        ui.output_error(message)
        raise ValueError(message)


class NamedGroup(TypedDict):
    r"""
    - Contains a name & regular expression used to extract data from some text

    - Used as an _input_ to [extract_values](../ppa_tools.text/#extract_values)

    - The example below shows the dictionary structure in YAML

    - See the ___required___ keys underneath the example

    !!! example

        ```yaml linenums="1"
        name: username
        value: your username is (\w+)
        ```

    !!! danger "Required Keys"

        __name__: The name to give the extracted value.

        __expression__: A regular expression used to find the value in the text.

    """

    name: str
    expression: str


class FindReplace(TypedDict):
    find: str
    replace: str


def join(items: List[Union[str, int, float]], delimiter: str = "") -> str:
    """Join items of text together.

    !!! question "Minimum Plugin Version: 6.5.0"

    ??? info "Input"

        - __items:__ one or more pieces of text

        - __delimiter:__ optional delimiter to use (defaults to no delimiter)

    ??? success "Output"

        New text with each string in **items** joined together using **delimiter**.

    ??? example "Example 1 - No Delimiter"

        ```YAML
        - ppa_tools.text.join:
            items:
              - these
              - will
              - be
              - joined
          save: output
        ```

        The output will be:

        ```YAML
        thesewillbejoined
        ```

    ??? example "Example 2 - Custom Delimiter"

        ```YAML
        - ppa_tools.text.join:
            items:
              - 1
              - 2
              - 3
            delimiter: ", "
          save: output
        ```

        The output will be:

        ```YAML
        1, 2, 3
        ```
    """
    return delimiter.join(list(map(str, items)))


def replace(text: str, find: str, replace: str):
    """Replace a character or pattern in some text.

    !!! question "Minimum Plugin Version: 6.5.0"

    ??? info "Input"

        - __text:__ text to perform the replace on

        - __find:__ character or pattern to find inside **text**

        - __replace:__ new text to replace any occurrences of **find** with

    ??? success "Output"

        New text with all occurrences of **find** replaced by **replace**.

    ??? example

        ```YAML
        - ppa_tools.text.replace:
            text: This is an example string
            find: is
            replace: was
          save: output
        ```

        The output will be:

        ```YAML
        This was an example string
        ```
    """
    return text.replace(find, replace)


def replace_many(text: str, criteria: List[FindReplace]) -> str:
    """Replace multiple characters or patterns in some text.

    This action runs the [replace](#replace) action multiple times.

    !!! question "Minimum Plugin Version: 6.5.0"

    ??? info "Input"

        - __text:__ text to perform the replace on

        - __criteria:__ a list of dictionaries containing **find** & **replace** values

    ??? success "Output"

        New text with all occurrences of each **find** replaced by each **replace**.

    ??? example

        ```YAML
        - ppa_tools.text.replace:
            text: This is a longer example string with multiple replace criteria
            criteria:
              - find: is
                replace: was
              - find: replace
                replace: replaced
          save: output
        ```

        The output will be:

        ```YAML
        This was a longer example string with multiple replaced criteria
        ```
    """
    for find_replace in criteria:
        text = replace(text, find_replace["find"], find_replace["replace"])
    return text


def split_lines(text: str, filter_empty: Optional[bool] = False):
    """Split a string into a list of strings based on new lines.

    !!! question "Minimum Plugin Version: 6.4.0"

    ??? info "Input"

        - __text:__ the string to split

        - __filter_empty:__ set to `true` to remove empty lines from the outputted list

    ??? success "Output"

        A list of strings.

    ??? example

        Splitting on one or more new line characters:

        ```YAML
        - ppa_tools.text.split_lines:
            text: >
              London
              Paris
              New York
        save: split_output
        ```

        The output will be:

        ```YAML
        - London
        - Paris
        - New York
        ```
    """
    lines = text.replace("\r", "").split("\n")
    if filter_empty is True:
        return [line for line in lines if len(line) > 0]
    return lines


def extract_values(values: Union[List[NamedGroup], NamedGroup], text: str) -> List[Dict[str, str]]:
    """Find & extract one or more named values from some text using regular expressions.

    This action supports both single & multiple lines of text.

    !!! question "Minimum Plugin Version: 1.3.0"

    !!! tip "Regular Expressions"

        The following [regex cheat-sheet](https://cheatography.com/davechild/cheat-sheets/regular-expressions/) & [visualiser](https://regexper.com/) can help you get started with expressions.

    ??? info "Input"

        - __values:__ a single or list of [NamedGroup](../inputs_outputs/#namedgroup) dictionaries

        - __text:__ the text to find & extract values from

    ??? success "Output"

        A list of dictionaries containing any matches found in the text.

        Each dictionary's _key_ is the `name` of the [NamedGroup](../inputs_outputs/#namedgroup), & its value is the matching text.

    ??? example

        Finding all usernames from a failed SSH logon message & saving the matches as `usernames`.

        ```YAML linenums="1"
        - ppa_tools.text.extract_values:
            values:
              name: username
              expression: password for (.*) from
            text: >
              2020-08-11T08:19:40.463248+00:00 machine-name sshd[10349]: Failed password for user_1 from 1.2.3.4 port 58041 ssh2
              2020-08-12T13:24:07.463248+00:00 machine-name sshd[10349]: Failed password for user_2 from 2.3.4.5 port 58041 ssh2
          save: usernames
        ```

        The contents of the new `usernames` variable will be:

        ```YAML linenums="1"
        - username: user_1
          username: user_2
        ```
    """

    def _search(text, expression):
        if match := re.search(expression, text):
            groups = match.groups()
            if len(groups) > 1:
                raise AmbiguousExpression(f"Expression '{expression}' found more than one match on line '{line}'.")
            return groups[0]
        logger.debug(f"no match for '{expression}' in '{repr(line)}'")
        return None

    values = [values] if not isinstance(values, list) else values

    invalid_expressions = []
    for named_group in values:
        try:
            re.compile(named_group["expression"])
        except re.error:
            invalid_expressions.append(named_group["expression"])
    if invalid_expressions:
        ui.output_error("One or more expressions supplied to search_text is invalid. Please view the logs for more information.")
        raise InvalidExpression("The following expressions are invalid: {}".format(", ".join(invalid_expressions)))

    matches = []
    for line in [line for line in text.split("\n") if line]:
        matches_on_line = {}
        for named_group in values:
            if result := _search(line, named_group["expression"]):
                matches_on_line[named_group["name"]] = result
        if not matches_on_line:
            logger.info(f"no matches on line '{line}'")
            # Ignore the line if no matches were found.
            continue
        matches.append(matches_on_line)
    return matches


def get_line_number(text: str, line: str) -> int:
    """Get the line number of the supplied line from some text.

    !!! question "Minimum Plugin Version: 5.9.0"

    ??? info "Input"

        - __text:__ the text to look in

        - __line:__ the line to look for

    !!! tip "Line Numbers start at 0"

    ??? success "Output"

        The line number (zero-based).

    ??? example

        Splitting on one or more new line characters:

        ```YAML
        - ppa_tools.text.get_line_number:
            text: >
              Line one
              Line two
              Line three
            line: Line two
          save: line_number
        ```

        The output will be:

        ```YAML
        1
        ```
    """
    lines = split_lines(text)
    if lines.count(line) == 0:
        message = "Could not find line in the supplied text."
        ui.output_error(message)
        raise ValueError(message)
    elif lines.count(line) > 1:
        message = "More than one line was found in the supplied text."
        ui.output_error(message)
        raise ValueError(message)
    return lines.index(line)


def extract_lines(text: str, start_line: Optional[Union[str, int]] = None, end_line: Optional[Union[str, int]] = None,) -> str:
    """Extract a number of lines from some text.

    !!! question "Minimum Plugin Version: 5.9.0"

    ??? info "Input"

        - __text:__ the text to look in

        - __start_line:__ optional line number to start at (defaults to the first line)

        - __end_line:__ optional line number to finish on (defaults to extracting all text after **start_line**)

    !!! tip "Line Numbers start at 0"

    ??? success "Output"

        The text found between the **start_line** & **end_line**.

    ??? example

        ```YAML
        - ppa_tools.text.extract_lines:
            text: >
              Line one
              Line two
              Line three
            start_line: 1
          save: extracted_lines
        ```

        The output will be:

        ```YAML
        Line two
        Line three
        ```
    """

    split_text = split_lines(text)
    length = len(split_text)

    # Default line numbers if either/or is not provided
    start_line = start_line if start_line is not None else 0
    end_line = end_line if end_line is not None else length

    # If supplied they may be strings (easy for people to do in YAML)
    if not isinstance(start_line, int):
        start_line = _to_int(start_line)
    if not isinstance(end_line, int):
        end_line = _to_int(end_line)

    if start_line > end_line:
        message = f"Start line {start_line} is greater than end line {end_line}."
        ui.output_error(message)
        raise ValueError(message)

    if start_line is not None and (start_line + 1) > length:
        message = f"Cannot get starting line number {start_line} from file with {length} lines."
        ui.output_error(message)
        raise IndexError(message)

    elif end_line is not None and end_line > length:
        message = f"Cannot get final line number {end_line} from file with {length} lines."
        ui.output_error(message)
        raise IndexError(message)

    return "\n".join(split_lines(text)[start_line:end_line])


def append(base: str, new: str) -> str:
    """Append one piece of text to the end of another.

    !!! question "Minimum Plugin Version: 5.9.0"

    ??? info "Input"

        - __base:__ the base text to append to

        - __new:__ the new text to append

    ??? success "Output"

        Text containing **base** with **new** appended to the end.

    ??? example

        ```YAML
        - ppa_tools.text.append:
            base: London
            new: Paris
          save: new_copy
        ```

        The output will be:

        ```YAML
        London
        Paris
        ```
    """
    lines = split_lines(base)
    lines.extend(split_lines(new))
    return "\n".join(lines)


def replace_lines(base: str, insert: str, start_line: Union[str, int], end_line: Union[str, int]) -> str:
    """Replace a number of lines inside some text.

    !!! question "Minimum Plugin Version: 5.9.0"

    ??? info "Input"

        - __base:__ the text to perform the replacement on

        - __insert:__ some text to insert

        - __start_line:__ optional line number to start at (defaults to the first line)

        - __end_line:__ optional line number to finish on (defaults to extracting all text after **start_line**)

    !!! tip "Line Numbers start at 0"

    ??? success "Output"

        The text found between the **start_line** & **end_line**.

    ??? example

        ```YAML
        - ppa_tools.text.replace_lines:
            base: |
              London
              Paris
              New York
              Brussels
              Manchester
            insert: |
              Bruges
              San Francisco
            start_line: 1
            end_line: 3
          save: replaced_text
        ```

        The output will be:

        ```YAML
        London
        Bruges
        San Francisco
        Brussels
        Manchester
        ```
    """
    split_original = split_lines(base)
    length = len(split_original)

    # If supplied they may be strings (easy for people to do in YAML)
    if not isinstance(start_line, int):
        start_line = _to_int(start_line)
    if not isinstance(end_line, int):
        end_line = _to_int(end_line)

    if start_line > end_line:
        message = f"Start line {start_line} is greater than end line {end_line}."
        ui.output_error(message)
        raise ValueError(message)

    if (start_line + 1) > length:
        message = f"Cannot get starting line number {start_line} from file with {length} lines."
        ui.output_error(message)
        raise IndexError(message)

    if end_line > length:
        message = f"Cannot get final line number {end_line} from file with {length} lines."
        ui.output_error(message)
        raise IndexError(message)

    new_copy = split_original[:start_line]  # First portion of original
    new_copy.extend(split_lines(insert))  # New text to insert
    new_copy.extend(split_original[end_line:])  # Last portion of original
    return "\n".join(new_copy)
